﻿/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using GestureToolkit;

namespace GECko
{
    public class Canvas
    {
        /// <summary>
        /// Draws a set of gestures into a bitmap.
        /// The list *indexesToDraw* indicates the subset of gestures to draw.
        /// The list *indexesToHighlight* indicates which gestures should be marked differently (with red X bars).
        /// Gestures are drawn in a matrix format having *numCols* columns.
        /// Each gesture is drawn inside a bounding rectangle of *gestureSize* x *gestureSize* pixels using *strokeColor*.
        /// </summary>
        /// <param name="gestures"></param>
        /// <param name="gestureSize"></param>
        /// <param name="numGesturesPerRow"></param>
        /// <returns></returns>
        public static Bitmap CreateBitmapFromGestureSet(
            List<Gesture> gestures, List<int> indexesToDraw, List<int> indexesToHighlight, int numCols,
            int gestureSize, Color strokeColor, 
            bool showBoundingRectangle = false, string clusterIndex = "")
        {
            int GESTURE_SIZE = gestureSize;
            int SPACE_BETWEEN_GESTURES = 4;
            int numRows = (int)Math.Ceiling(indexesToDraw.Count * 1.0 / numCols);

            // create bitmap
            Bitmap bitmap = new Bitmap(
                SPACE_BETWEEN_GESTURES + (GESTURE_SIZE + SPACE_BETWEEN_GESTURES) * numCols + GESTURE_SIZE, 
                SPACE_BETWEEN_GESTURES + (GESTURE_SIZE + SPACE_BETWEEN_GESTURES) * numRows
            );

            // draw gestures
            Graphics graphics = null;
            Font font = null;
            try
            {
                graphics = Graphics.FromImage(bitmap);
                font = new Font("Tahoma", 6.5f);

                // draw each gesture sample
                for (int i = 0; i < indexesToDraw.Count; i++)
                {
                    Gesture gesture = gestures[indexesToDraw[i]];
                    using (Bitmap gestureBitmap = Canvas.GestureToBitmap(gesture, GESTURE_SIZE, strokeColor))
                    {
                        // compute the row and column
                        int indexRow = i / numCols;
                        int indexCol = i % numCols;

                        // draw bitmap
                        int left = SPACE_BETWEEN_GESTURES + indexCol * (GESTURE_SIZE + SPACE_BETWEEN_GESTURES);
                        int top = SPACE_BETWEEN_GESTURES + indexRow * (GESTURE_SIZE + SPACE_BETWEEN_GESTURES);
                        graphics.DrawImage(gestureBitmap, left, top);

                        // draw gesture name
                        string text = gesture.Name;
                        text = text.Substring(0, Math.Min(text.Length, 4)) + String.Format("{0}", i + 1);
                        SizeF textSize = graphics.MeasureString(text, font);
                        graphics.DrawString(
                            text, font, Brushes.Gray,
                            left + gestureBitmap.Width - textSize.Width,
                            top + 1
                        );

                        // draw the number of strokes in the bottom right corner
                        text = String.Format("{0}", gesture.GetNumberOfStrokes());
                        textSize = graphics.MeasureString(text, font);
                        graphics.DrawString(
                            text, font, Brushes.Gray,
                            left + gestureBitmap.Width - textSize.Width,
                            top + gestureBitmap.Height - textSize.Height
                        );

                        // is highlighted?
                        if (indexesToHighlight.IndexOf(indexesToDraw[i]) != -1)
                        {
                            Pen pen = new Pen(Color.Red, 2);
                            graphics.DrawLine(pen, left, top, left + gestureBitmap.Width, top + gestureBitmap.Height);
                            graphics.DrawLine(pen, left + gestureBitmap.Width, top, left, top + gestureBitmap.Height);
                            pen.Dispose();
                        }
                    }
                }

                // show a rectangle around the cluster?
                if (showBoundingRectangle)
                    graphics.DrawRectangle(Pens.White, new Rectangle(0, 0, bitmap.Width - 1, bitmap.Height - 1));

                // draw the cluster index?
                if (clusterIndex.Length > 0)
                {
                    string text = clusterIndex;
                    SizeF textSize = graphics.MeasureString(text, font);
                    graphics.DrawString(
                      text, font, Brushes.Gray,
                      bitmap.Width - textSize.Width - 3, 2
                    );
                }
            }
            finally
            {
                if (font != null)
                    font.Dispose();
                if (graphics != null)
                    graphics.Dispose();
            }

            return bitmap;
        }

        /// <summary>
        /// Draws a scaled version of the *gesture* into a bitmap of size *width* x *height* with *strokeColor*.
        /// If *showPartial*, then only the first points up to *partialIndex* are displayed (used to animate the gesture).
        /// </summary>
        /// <param name="gesture"></param>
        /// <returns></returns>
        public static Bitmap GestureToBitmap(
            Gesture gesture, int gestureSize, Color strokeColor, 
            bool showPartial = false, int partialIndex = 0)
        {
            Bitmap bitmap = new Bitmap(gestureSize, gestureSize);

            Graphics graphics = null;
            Pen penStrokes = null;
            try
            {
                // clear graphics to white
                graphics = Graphics.FromImage(bitmap);
                graphics.Clear(Color.White);
                graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;  // smooth lines
                graphics.InterpolationMode = InterpolationMode.Bicubic;

                // draw the gesture
                System.Drawing.Point previousPoint = System.Drawing.Point.Empty;
                List<System.Drawing.Point> strokeStartingPoints = new List<System.Drawing.Point>();
                int n = showPartial ? partialIndex : gesture.m_normalizedPoints.Length;
                float strokePartialLength = 0, strokeLength = 0; // length and partial length of the current stroke (for the fading effect)
                int indexStroke = -1;
                for (int i = 0; i < n; i++)
                {
                    System.Drawing.Point currentPoint = RescaleNormalizedPoint(gesture.m_normalizedPoints[i], gestureSize, gestureSize); 

                    // starting point for a new stroke?
                    if (i == 0 || (gesture.m_strokeIndex[i] != gesture.m_strokeIndex[i - 1]))
                    {
                        strokeStartingPoints.Add(currentPoint);

                        // stroke index
                        indexStroke++;

                        // get the length of the current stroke
                        strokeLength = 0;
                        int strokeEnd = i + 1;
                        while (strokeEnd < n && gesture.m_strokeIndex[strokeEnd] == gesture.m_strokeIndex[strokeEnd - 1])
                        {
                            strokeLength += Measurements.EuclideanDistance(gesture.m_normalizedPoints[strokeEnd - 1], gesture.m_normalizedPoints[strokeEnd]);
                            strokeEnd++;
                        }
                        strokePartialLength = 0;
                    }
                    else // same stroke
                    {
                        // draw with stroke color fading away to suggest execution direction
                        strokePartialLength += Measurements.EuclideanDistance(gesture.m_normalizedPoints[i - 1], gesture.m_normalizedPoints[i]);
                        double t = strokeLength < 10e-3 ? 0.0 : Math.Max(0, Math.Min(1.0, strokePartialLength / strokeLength));
                        Color startColor = GetColorForStrokeIndex(indexStroke, 20);
                        Color endColor = Color.FromArgb(240, 240, 240);
                        Color color = Color.FromArgb((int)((1 - t) * startColor.R + t * endColor.R), (int)((1 - t) * startColor.G + t * endColor.G), (int)((1 - t) * startColor.B + t * endColor.B));
                        using (Pen pen = new Pen(color, 3.5f))
                        {
                            pen.EndCap = LineCap.Round;
                            pen.StartCap = LineCap.Round;
                            graphics.DrawLine(pen, previousPoint, currentPoint);
                        }
                    }

                    // store the current point
                    previousPoint = currentPoint;
                }

                // draw  the starting point of each stroke in a distinctive manner: large dot and different color
                for(int i = 0; i < strokeStartingPoints.Count; i++)
                {
                    int radius = Math.Min(5, (int)(gestureSize * 0.075));
                    Rectangle pointRectangle = new Rectangle(strokeStartingPoints[i].X - radius, strokeStartingPoints[i].Y - radius, 2 * radius, 2 * radius);
                    SolidBrush brushStartingPoint = new SolidBrush(GetColorForStrokeIndex(i, strokeStartingPoints.Count));
                    
                    graphics.FillEllipse(brushStartingPoint, pointRectangle);
                    graphics.DrawEllipse(Pens.Black, pointRectangle);
                    brushStartingPoint.Dispose();
                }
            }
            finally
            {
                if (penStrokes != null)
                    penStrokes.Dispose();
                if (graphics != null)
                    graphics.Dispose();
            }

            return bitmap;
        }

        /// <summary>
        /// Returns a distinct color for each stroke index
        /// </summary>
        /// <param name="strokeIndex"></param>
        /// <returns></returns>
        private static Color GetColorForStrokeIndex(int strokeIndex, int numStrokes)
        {
            // predefined colors
            Color[] colors = new Color[]
            {
                Color.Red,
                Color.Gray,
                Color.Green,
                Color.Navy,
                Color.Orange,
                Color.Magenta,
                Color.Turquoise,
            };

            if (strokeIndex < colors.Length)
                return colors[strokeIndex];
            else
            {
                int grayValue = numStrokes == 1 ? 0 : (int)(200.0 * strokeIndex / (numStrokes - 1));
                return Color.FromArgb(grayValue, grayValue, grayValue);
            }
        }

        /// <summary>
        /// Converts a normalized gesture point from [-1,1] x [-1,1] to [0,width] x [0,height]
        /// </summary>
        /// <param name="point"></param>
        /// <param name="width"></param>
        /// <param name="height"></param>
        /// <returns></returns>
        private static System.Drawing.Point RescaleNormalizedPoint(GestureToolkit.Point point, int width, int height)
        {
            return new System.Drawing.Point(
                        (int)((point.X + 1.0) / 2.0 * width),
                        (int)((point.Y + 1.0) / 2.0 * height)
                    );
        }
    }
}
