﻿/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using GestureToolkit;

namespace GECko
{
    public class ClustersGUI
    {
        /// <summary>
        /// list of gestures subjected to clustering analysis
        /// </summary>
        private List<Gesture> gestures = null;

        /// <summary>
        /// list of clusters and associated data (picture boxes and user selected samples)
        /// </summary>
        private List<Cluster> clusters = null;
        private List<PictureBox> pbClusters = null;
        private List<Cluster> selectedSamples = null;

        /// <summary>
        /// the GUI elements for the main form
        /// </summary>
        private Panel parentControl = null;                                  // the container for the cluster picture boxes
        private int indexClusterForAutoScroll = 0;                           // the cluster index that should be in view 
        private Button btnRemove = null, btnSplit = null, btnMerge = null, btnSort = null;
        private Label lblClustersStats = null;                               // displays stats information on the current clustering
        private PictureBox pbPlayGesture = null;                             // plays the gesture
        
        /// <summary>
        /// gesture drawing
        /// </summary>
        public static int GESTURE_SIZE = 50;         // size of the bouding box for each gesture sample in pixels
        private static int GESTURES_PER_ROW = 15;    // display 15 gesture samples in a row

        /// <summary>
        /// Creates a new object for the GUI dealing with displaying and manual editing of gesture clusters
        /// </summary>
        /// <param name="gestures"></param>
        public ClustersGUI(Panel parentControl, Button btnRemove, Button btnMerge, Button btnSplit, Button btnSortClusters, Label lblClustersStats, PictureBox pbPlayGesture)
        {
            this.parentControl = parentControl;
            this.btnRemove = btnRemove;
            this.btnRemove.Click += new EventHandler(btnRemove_Click);
            this.btnMerge = btnMerge;
            this.btnMerge.Click += new EventHandler(btnMerge_Click);
            this.btnSplit = btnSplit;
            this.btnSplit.Click += new EventHandler(btnSplit_Click);
            this.btnSort = btnSortClusters;
            this.btnSort.Click += new EventHandler(btnSort_Click);
            this.lblClustersStats = lblClustersStats;
            this.pbPlayGesture = pbPlayGesture;
            this.pbPlayGesture.Click += new EventHandler(pbPlayGesture_Click);

            btnSort.Enabled = true;
        }

        /// <summary>
        /// Sets gesture samples and clusters
        /// </summary>
        /// <param name="gestures"></param>
        /// <param name="clusters"></param>
        public void SetGestureData(List<Gesture> gestures, List<Cluster> clusters)
        {
            this.gestures = gestures;
            this.clusters = clusters;
            SortClusters(true);
            pbPlayGesture.Visible = false;
            indexClusterForAutoScroll = 0;
        }

        /// <summary>
        /// Get clusters
        /// </summary>
        /// <returns></returns>
        public List<Cluster> GetClusters()
        {
            return this.clusters;
        }

        /// <summary>
        /// Displays gesture clusters as picture boxes
        /// </summary>
        public void DisplayClusters()
        {
            // hide the play gesture control
            pbPlayGesture.Visible = false;

            // no gesture samples selected yet
            this.selectedSamples = new List<Cluster>();
            for (int i = 0; i < clusters.Count; i++)
               selectedSamples.Add(new Cluster());
            
            // make sure that existing picture boxes (no longer used) are being disposed correctly
            FreeMemory();

            // display each cluster inside its own picture box
            pbClusters = new List<PictureBox>();
            parentControl.SuspendLayout();
            for (int i = 0; i < clusters.Count; i++)
            {
                PictureBox pb = new PictureBox();

                pb.Name = String.Format("Cluster{0}", i);
                pb.Location = new System.Drawing.Point(2, i == 0 ? 3 : pbClusters[i - 1].Bottom + 15);
                pb.SizeMode = PictureBoxSizeMode.AutoSize;
                pb.BackColor = Color.Empty;
                pb.Cursor = Cursors.Hand;   // show user an indication to click and select
                pb.Tag = i;                 // needed in order to easily identify the cluster index

                pb.Image = Canvas.CreateBitmapFromGestureSet(
                    gestures,
                    clusters[i], selectedSamples[i],
                    GESTURES_PER_ROW, GESTURE_SIZE, Color.Gray,
                    true, (i + 1).ToString()
                );

                pb.MouseDown += new MouseEventHandler(pbCluster_MouseDown);
                pbClusters.Add(pb);
                parentControl.Controls.Add(pb);
            }
            parentControl.ResumeLayout(false);
            parentControl.PerformLayout();

            // autoscroll the container in order to bring the *indexClusterForAutoScroll* into view
            parentControl.AutoScrollPosition = new System.Drawing.Point(0, pbClusters[indexClusterForAutoScroll].Top - parentControl.Height / 3);
            
            // update clusters statistics
            float totalSamples = 0;
            for (int i = 0; i < clusters.Count; i++)
                totalSamples += clusters[i].Count;
            lblClustersStats.Text = String.Format("Clusters: {0}  |  Total samples: {1}  |  Average cluster size: {2:.0}", clusters.Count, totalSamples, totalSamples / (clusters.Count * 1.0f));
        }

        /// <summary>
        /// Catch mouse clicks on picture boxes to select/unselect gesture samples and select/unselect clusters for editing operations
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void pbCluster_MouseDown(object sender, MouseEventArgs e)
        {
            // get currently selected picture box and cluster
            PictureBox pb = (PictureBox)sender;
            int indexCluster = (int)(pb.Tag);
            Cluster cluster = clusters[indexCluster];

            // was it a click on top of a gesture sample?
            int row = (e.Location.Y - 4) / (GESTURE_SIZE + 4);      // see the Canvas class for the meaning of 4
            int col = (e.Location.X - 4) / (GESTURE_SIZE + 4);
            int indexDisplayedGesture = row * GESTURES_PER_ROW + col;
            if (col < GESTURES_PER_ROW && indexDisplayedGesture < cluster.Count)
            {
                int indexGesture = cluster[indexDisplayedGesture];
                switch (e.Button)
                {
                    case MouseButtons.Left:
                        // select/unselect the gesture sample
                        if (selectedSamples[indexCluster].IndexOf(indexGesture) == -1)
                            selectedSamples[indexCluster].Add(indexGesture);
                        else selectedSamples[indexCluster].Remove(indexGesture);
                        selectedSamples[indexCluster].Sort();   // make sure indexes are sorted

                        // update the picture box image of the cluster in order to display the new selection
                        pb.Image.Dispose();
                        pb.Image = Canvas.CreateBitmapFromGestureSet(
                            gestures, cluster, selectedSamples[indexCluster],
                            GESTURES_PER_ROW, GESTURE_SIZE, Color.Gray,
                            true, (indexCluster + 1).ToString()
                        );
                        pb.Refresh();
                        break;
                    case MouseButtons.Right:
                        PlayGesture(gestures[indexGesture], new System.Drawing.Point(Math.Max(2, col < 10 ? e.Location.X + pb.Left : e.Location.X + pb.Left - 150), Math.Max(2, pb.Top < 300 ? e.Location.Y + pb.Top - 20 : e.Location.Y + pb.Top - 150)));
                        break;
                }
            }
            else
            {
                // if mouse click was outside any gesture sample then select / unselect the entire cluster
                if (IsClusterSelected(indexCluster))
                    UnselectCluster(indexCluster);
                else SelectCluster(indexCluster);
            }

            // enable / disable the Remove and Split buttons
            bool enabled = false;
            for (int i = 0; i < selectedSamples.Count; i++)
                if (selectedSamples[i].Count > 0)
                {
                    enabled = true;
                    break;
                }
            btnRemove.Enabled = enabled;
            btnSplit.Enabled = enabled;

            // enable / disable the Merge clusters button
            btnMerge.Enabled = GetSelectedClusters().Count >= 2 ? true : false;
        }

        private bool IsClusterSelected(int index)
        {
            return pbClusters[index].BackColor == Color.LightSlateGray ? true : false;
        }

        private void SelectCluster(int index)
        {
            pbClusters[index].BackColor = Color.LightSlateGray;
        }

        private void UnselectCluster(int index)
        {
            pbClusters[index].BackColor = Color.Empty;
        }

        /// <summary>
        /// Plays *gesture* inside a picture box
        /// </summary>
        /// <param name="gesture"></param>
        private bool stopCurrentGesturePlay = false;
        private void PlayGesture(Gesture gesture, System.Drawing.Point location)
        {
            stopCurrentGesturePlay = false;
            pbPlayGesture.Location = location;
            pbPlayGesture.Visible = true;
            using (Graphics graphics = Graphics.FromImage(pbPlayGesture.Image))
            {
                Font font = new Font("Tahoma", 8);
                for (int i = 0; i < gesture.m_normalizedPoints.Length && !stopCurrentGesturePlay; i += 3)
                {
                    using (Bitmap bitmap = Canvas.GestureToBitmap(gesture, pbPlayGesture.Image.Width, Color.Gray, true, i))
                    {
                        graphics.DrawImage(bitmap, new PointF(0, 0));
                        graphics.DrawString(gesture.Name, font, Brushes.Gray, new PointF(2, 1));
                        graphics.DrawString("click to close", font, Brushes.Gray, new PointF(2, pbPlayGesture.Height - 18));
                    }
                    pbPlayGesture.Refresh();
                    Application.DoEvents();
                    System.Threading.Thread.Sleep(70);
                }
                font.Dispose();
            }
            stopCurrentGesturePlay = true;
        }

        /// <summary>
        /// Hide the playing gesture on click
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void pbPlayGesture_Click(object sender, EventArgs e)
        {
            stopCurrentGesturePlay = true;
            pbPlayGesture.Visible = false;
        }

        /// <summary>
        /// Frees the memory occupied by the list of picture boxes displaying clusters
        /// </summary>
        /// <param name="pictureBoxes"></param>
        public void FreeMemory()
        {
            if (pbClusters == null)
                return;

            for (int i = 0; i < pbClusters.Count; i++)
            {
                parentControl.Controls.Remove(pbClusters[i]);
                try
                {
                    if (pbClusters[i].Image != null)
                    {
                        pbClusters[i].Image.Dispose();
                        pbClusters[i].Image = null;
                    }
                }
                catch { }
                pbClusters[i].Dispose();
            }
        }

        /// <summary>
        /// Returns a list of all selected clusters
        /// </summary>
        /// <returns></returns>
        private List<int> GetSelectedClusters()
        {
            List<int> selectedClusters = new List<int>();
            for (int i = 0; i < clusters.Count; i++)
                if (IsClusterSelected(i))
                    selectedClusters.Add(i);
            return selectedClusters;
        }

        /// <summary>
        /// Make sure the gesture indexes in each cluster are sorted
        /// </summary>
        private void SortClusters(bool sortAdditionallyByName = false)
        {
            // sort indexes inside each cluster
            foreach (Cluster cluster in clusters)
                cluster.Sort();

            // sort clusters by gesture name
            if (sortAdditionallyByName)
                clusters.Sort(CompareClusters);
        }

        /// <summary>
        /// Sorts clusters
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void btnSort_Click(object sender, EventArgs e)
        {
            SortClusters(true);
            DisplayClusters();
        }

        /// <summary>
        /// Comparison function for ordering the clusters list.
        /// Returns < 0 if a < b
        /// Returns = 0 if a = b
        /// Returns > 0 if a > b
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        private int CompareClusters(Cluster a, Cluster b)
        {
            // sort clusters by gesture name (ASC)
            int stringComparison = string.Compare(gestures[a[0]].Name, gestures[b[0]].Name);
            if (stringComparison == 0)
            {
                // if gesture name is the same then sort clusters by the number of strokes (ASC)
                int numStrokesA = gestures[a[0]].GetNumberOfStrokes();
                int numStrokesB = gestures[b[0]].GetNumberOfStrokes();
                if (numStrokesA == numStrokesB)
                {
                    // sort clusters by their size (DESC, larger clusters go first)
                    return -a.Count.CompareTo(b.Count);
                }
                else return numStrokesA.CompareTo(numStrokesB);
            }
            else return stringComparison;
        }

        /// <summary>
        /// Remove selected samples from the current cluster and form a new cluster
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void btnRemove_Click(object sender, EventArgs e)
        {
            // create list of new clusters
            List<Cluster> newClusters = new List<Cluster>();

            // remove samples
            indexClusterForAutoScroll = 0;
            for (int i = 0; i < clusters.Count; i++)
            {
                // create a new cluster that will contain removed samples from this cluster
                Cluster newCluster = new Cluster();
                foreach (int index in selectedSamples[i])
                {
                    newCluster.Add(index);
                    clusters[i].Remove(index);
                }

                // add both original and new cluster to the new list of clusters
                if (clusters[i].Count > 0)
                    newClusters.Add(clusters[i].Clone());
                if (newCluster.Count > 0)
                {
                    newClusters.Add(newCluster);
                    indexClusterForAutoScroll = newClusters.Count - 1;
                }
            }

            // display new clusters
            this.clusters = newClusters;
            SortClusters();
            DisplayClusters();

            // disable remove and split buttons
            btnSplit.Enabled = false;
            btnRemove.Enabled = false;
            pbPlayGesture.Visible = false;
        }

        /// <summary>
        /// Merge selected clusters into a single cluster
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void btnMerge_Click(object sender, EventArgs e)
        {
            // create list of new clusters
            List<Cluster> newClusters = new List<Cluster>();

            // get selected clusters
            List<int> selectedClustersIndexes = GetSelectedClusters();
            if (selectedClustersIndexes.Count == 0)
                return;
            selectedClustersIndexes.Sort();

            // merge clusters
            indexClusterForAutoScroll = 0;
            for (int i = 0; i < clusters.Count; i++)
            {
                if (i == selectedClustersIndexes[0])
                {
                    Cluster newCluster = new Cluster();
                    foreach (int indexCluster in selectedClustersIndexes)
                        newCluster.AddRange(clusters[indexCluster].Clone());
                    newClusters.Add(newCluster);
                    indexClusterForAutoScroll = newClusters.Count - 1;
                }
                else
                {
                    if (selectedClustersIndexes.IndexOf(i) == -1)
                        newClusters.Add(clusters[i].Clone());
                }
            }

            // display new clusters
            this.clusters = newClusters;
            SortClusters();
            DisplayClusters();

            // disable merge button
            btnMerge.Enabled = false;
            pbPlayGesture.Visible = false;
        }

        /// <summary>
        /// Split clusters at first selection point
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void btnSplit_Click(object sender, EventArgs e)
        {
            // create list of new clusters
            List<Cluster> newClusters = new List<Cluster>();

            // split clusters
            indexClusterForAutoScroll = 0;
            for (int i = 0; i < clusters.Count; i++)
            {
                if (selectedSamples[i].Count > 0)
                {
                    // create two new clusters that will contain samples from this cluster
                    Cluster newCluster1 = new Cluster();
                    Cluster newCluster2 = new Cluster();
                    int thresholdIndex = selectedSamples[i][0];
                    bool addToCluster1 = true;
                    foreach (int index in clusters[i])
                    {
                        if (index == thresholdIndex)
                            addToCluster1 = false;
                        if (addToCluster1)
                            newCluster1.Add(index);
                        else newCluster2.Add(index);
                    }

                    if (newCluster1.Count > 0)
                        newClusters.Add(newCluster1);
                    if (newCluster2.Count > 0)
                        newClusters.Add(newCluster2);
                    indexClusterForAutoScroll = newClusters.Count - 1;
                }
                else newClusters.Add(clusters[i].Clone());
            }

            // display new clusters
            this.clusters = newClusters;
            SortClusters();
            DisplayClusters();

            // disable remove and split buttons
            btnSplit.Enabled = false;
            btnRemove.Enabled = false;
            pbPlayGesture.Visible = false;
        }
    }
}
