﻿/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.IO;
using GestureToolkit;

namespace GECko
{
    public partial class FormMain : Form
    {
        /// <summary>
        /// List of gestures that are subjected to clustering analysis
        /// </summary>
        private List<Gesture> gestures = null;    

        /// <summary>
        /// Complete list of gestures loaded from files
        /// 1) single-user scenario: all gestures are being clustered ==> allGestures = gestures
        /// 2) multiple-user scenario: only representative gestures are being clustered ==> gestures contains one representative gesture per cluster and user
        /// </summary>
        private List<Gesture> allGestures = null;

        /// <summary>
        /// GUI component for displaying and editing gesture clusters
        /// </summary>
        private ClustersGUI clustersGUI = null;   
        
        #region form constructor & shut down

        /// <summary>
        /// Performs initialization of GUI components
        /// </summary>
        public FormMain()
        {
            InitializeComponent();

            // initial folder to load up gesture sets and save results
            folderBrowserDialog.SelectedPath = Application.StartupPath;
            
            // hide results tab pages for now
            tabControlResults.TabPages.Remove(tabPageGestureSet);
            tabControlResults.TabPages.Remove(tabPageClusters);

            // init mode for single-user analysis
            panelAutoClusterSingleUser.Enabled = false;
            btnSaveSingleUser.Enabled = false;

            // init mode for multiple-user analysis
            panelAutoClusterMultipleUsers.Enabled = false;
            btnSaveMultipleUsers.Enabled = false;

            // init mode for cluster editing operations
            btnRemove.Enabled = false;
            btnSplit.Enabled = false; 
            btnMerge.Enabled = false;
            btnSortClusters.Enabled = false;
            pbPlayGesture.Visible = false;
            pbPlayGesture.Image = new Bitmap(pbPlayGesture.Width, pbPlayGesture.Height);

            // create cluster GUI
            clustersGUI = new ClustersGUI(
                panelClusters, btnRemove, btnMerge, btnSplit, btnSortClusters, lblClustersStats, pbPlayGesture
            );
        }
           
        /// <summary>
        /// Shutting down: free resources
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // clean up GUI resources used for editing clusters
            if (clustersGUI != null)
                clustersGUI.FreeMemory();
        }

        #endregion

        #region load gesture data

        /// <summary>
        /// Loads and displays a gesture set (single-user scenario)
        /// </summary>
        private void btnLoadData(object sender, EventArgs e)
        {
            if (sender == btnLoadGesturesSingleUser)
                folderBrowserDialog.Description = "Select a folder containing gesture samples for a given participant (xml files, one gesture per file).";
            else folderBrowserDialog.Description = "Select a folder containing participant subfolders (one subfolder per participant).";
            
            if (folderBrowserDialog.ShowDialog() == DialogResult.OK)
            {
                //folderBrowserDialog.SelectedPath = @"C:\Documents and Settings\radu\Desktop\MultiStroke\GECKo\GECko\bin\Debug\mmg-dataset\22-stylus-MEDIUM";
                     
                // read gestures
                if (sender == btnLoadGesturesSingleUser)
                    LoadGesturesSingleUser(folderBrowserDialog.SelectedPath);
                else
                    LoadGesturesMultipleUsers(folderBrowserDialog.SelectedPath);

                // display gestures
                List<int> indexes = new List<int>();
                for (int i = 0; i < gestures.Count; i++)
                    indexes.Add(i);
                if (pbGestureSet.Image != null)
                    pbGestureSet.Image.Dispose();
                pbGestureSet.Image = Canvas.CreateBitmapFromGestureSet(gestures, indexes, new List<int>(), 10, ClustersGUI.GESTURE_SIZE, Color.Gray);
                pbGestureSet.Refresh();

                // show the gesture set tab
                if (!tabControlResults.TabPages.Contains(tabPageGestureSet))
                    tabControlResults.TabPages.Add(tabPageGestureSet);
                if (tabControlResults.TabPages.Contains(tabPageClusters))
                    tabControlResults.TabPages.Remove(tabPageClusters);

                // enable/disable next step buttons
                if (gestures.Count > 0)
                {
                    if (sender == btnLoadGesturesSingleUser)
                    {
                        panelAutoClusterSingleUser.Enabled = true;
                        btnSaveSingleUser.Enabled = false;
                        panelAutoClusterMultipleUsers.Enabled = false;
                        btnSaveMultipleUsers.Enabled = false;
                    }
                    else
                    {
                        panelAutoClusterMultipleUsers.Enabled = true;
                        btnSaveMultipleUsers.Enabled = false;
                        panelAutoClusterSingleUser.Enabled = false;
                        btnSaveSingleUser.Enabled = false;
                    }
                }
            }
        }

        /// <summary>
        /// Reads all the gestures found under *baseFolder*
        /// </summary>
        /// <param name="folder"></param>
        /// <returns></returns>
        private void LoadGesturesSingleUser(string baseFolder)
        {
            // create list of gestures
            gestures = new List<Gesture>();     // for this scenario, *gestures* contains all loaded gestures
            string[] xmlFiles = Directory.GetFiles(baseFolder, "*.xml");
            foreach (string fileName in xmlFiles)
                gestures.Add(GestureIO.ReadGesture(fileName));

            // *allGestures* references the same list
            allGestures = gestures;
        }

        /// <summary>
        /// Reads all the gestures found under *baseFolder*.
        /// The *baseFolder* should contain subfolders representing different participants.
        /// </summary>
        /// <param name="baseFolder"></param>
        private void LoadGesturesMultipleUsers(string baseFolder)
        {
            // load data from multiple participants, each participant with its own folder
            string[] participants = Directory.GetDirectories(baseFolder);
            
            // create list of gestures
            gestures = new List<Gesture>();     // for this scenario, *gestures* contains representative gestures (one gesture per class and participant)
            allGestures = new List<Gesture>();  // also store all loaded gestures
            foreach (string participantFolder in participants)
            {
                // read cluster file
                string clusterFile = String.Format("{0}\\clusters-{1}.csv", participantFolder, participantFolder.Substring(participantFolder.LastIndexOf("\\") + 1));
                if (!File.Exists(clusterFile))
                {
                    MessageBox.Show(String.Format("File not found: \"{0}\".\r\nPlease run a single-user analysis for this user in order to produce the required file.", clusterFile), "File not found");
                    gestures.Clear();
                    return;
                }
                Dictionary<string, string> fileNameToClusterName = new Dictionary<string, string>();
                using (StreamReader reader = new StreamReader(clusterFile))
                {
                    // first line = header line
                    string line = reader.ReadLine();
                    // each next line = participant,gesture,numberOfStrokes,clusterName,gestureFileName
                    while ((line = reader.ReadLine()) != null)
                    {
                        string[] values = line.Split(new char[] { ',' });
                        fileNameToClusterName.Add(values[4], values[3]);
                    }
                }

                // read all xmlFiles (one xml file = one gesture sample)
                List<Gesture> participantGestures = new List<Gesture>();
                string[] xmlFiles = Directory.GetFiles(participantFolder, "*.xml");
                foreach (string fileName in xmlFiles)
                {
                    Gesture gesture = GestureIO.ReadGesture(fileName);
                    gesture.ClusterName = fileNameToClusterName[(new FileInfo(fileName)).Name];
                    participantGestures.Add(gesture);
                }

                // select representative gestures (one gesture per cluster)
                List<string> representativeClusters = new List<string>();
                foreach (Gesture gesture in participantGestures)
                {
                    if (!representativeClusters.Contains(gesture.ClusterName))
                    {
                        representativeClusters.Add(gesture.ClusterName);
                        gestures.Add(gesture);
                    }
                }

                // add this participant gestures to the list of all gestures
                allGestures.AddRange(participantGestures);
            }
        }

        #endregion

        #region automatic clustering

        /// <summary>
        /// Button action: run automatic clustering
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnAutomaticClustering(object sender, EventArgs e)
        {
            AutomaticClustering(sender);
        }

        /// <summary>
        /// Track bar action: on clustering threshold change, run the clustering procedure
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void trackBarClusteringThreshold_MouseUp(object sender, MouseEventArgs e)
        {
            AutomaticClustering(sender);
        }

        /// <summary>
        /// Runs the automatic clustering procedure for the gesture set and displays the result
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AutomaticClustering(object sender)
        {
            this.Cursor = Cursors.WaitCursor;

            // cluster the gesture set
            HierarchicalClustering clusteringAnalysis = new HierarchicalClustering();
            clustersGUI.FreeMemory();
            float clusteringThreshold = trackBarSingleUser.Value / 100.0f;
            clustersGUI.SetGestureData(gestures, clusteringAnalysis.Cluster(gestures.ToArray(), clusteringThreshold));

            this.Cursor = Cursors.Default;

            // show clusters 
            clustersGUI.DisplayClusters();
            if (!tabControlResults.TabPages.Contains(tabPageClusters))
                tabControlResults.TabPages.Add(tabPageClusters);
            tabControlResults.SelectedTab = tabPageClusters;

            // show save results button
            if (sender == btnClusterSingleUser || sender == trackBarSingleUser)
            {
                btnSaveSingleUser.Enabled = true;
            }
            else
            {
                btnSaveMultipleUsers.Enabled = true;
            }
        }

        #endregion
      
        #region save results

        /// <summary>
        /// Saves the clustering results as an Excel csv file (single-user scenario)
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnSaveSingleUser_Click(object sender, EventArgs e)
        {
            // store cluster name for each gesture
            List<Cluster> clusters = clustersGUI.GetClusters();
            for (int indexCluster = 0; indexCluster < clusters.Count; indexCluster++)
                foreach (int gestureIndex in clusters[indexCluster])
                    gestures[gestureIndex].ClusterName = String.Format("p{0}-c{1}", gestures[gestureIndex].ParticipantName, indexCluster + 1);

            // write output to file
            string outputFolder = folderBrowserDialog.SelectedPath;
            string participantName = folderBrowserDialog.SelectedPath.Substring(folderBrowserDialog.SelectedPath.LastIndexOf("\\") + 1);
            string outputFileClusters = String.Format("{0}\\clusters-{1}.csv", outputFolder, participantName);
            string outputFileAgreeementRates = String.Format("{0}\\agreementRates-{1}.csv", outputFolder, participantName);
            ClusterIO.WriteClustersToFile(outputFileClusters, gestures);
            
            // write agreement result
            ClusterIO.ComputeAgreementRates(outputFileClusters, outputFileAgreeementRates, gestures[0].ParticipantName);
            
            try
            {
                System.Diagnostics.Process.Start("explorer.exe", "/select," + outputFileClusters);
            }
            catch { }
        }

        /// <summary>
        /// Saves the clustering results as an Excel csv file (multiple-user scenario)
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnSaveMultipleUsers_Click(object sender, EventArgs e)
        {
            // use the representative gestures in order to merge clusters for multiple users
            Dictionary<string, string> oldClusterNameToNewClusterName = new Dictionary<string, string>();
            List<Cluster> clusters = clustersGUI.GetClusters();
            for (int indexCluster = 0; indexCluster < clusters.Count; indexCluster++)
            {
                string newClusterName = String.Format("c{0}", indexCluster + 1);
                foreach (int gestureIndex in clusters[indexCluster])
                {
                    // associate old cluster name (user-dependent as derived from the single-user analysis) to new cluster name 
                    oldClusterNameToNewClusterName.Add(gestures[gestureIndex].ClusterName, newClusterName);
                }
            }
                    
            // replace oldClusterName with newClusterName (user-independent as derived from the multi-user analysis)
            string[] oldnames = new string[allGestures.Count];
            for (int i = 0; i < allGestures.Count; i++)
            {
                oldnames[i] = allGestures[i].ClusterName;
                allGestures[i].ClusterName = oldClusterNameToNewClusterName[allGestures[i].ClusterName];
            }
                        
            // write output to file
            ClusterIO.WriteClustersToFile(folderBrowserDialog.SelectedPath + "\\clusters.csv", allGestures);

            // write agreement result
            ClusterIO.ComputeAgreementRates(
                folderBrowserDialog.SelectedPath + "\\clusters.csv",
                folderBrowserDialog.SelectedPath + "\\agreementRates.csv",
                ""
            );

            try
            {
                System.Diagnostics.Process.Start("explorer.exe", "/select," + folderBrowserDialog.SelectedPath + "\\clusters.csv");
            }
            catch { }

            for (int i = 0; i < allGestures.Count; i++)
                allGestures[i].ClusterName = oldnames[i];
        }

        #endregion
    }
}
