﻿/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.IO;
using System.Collections.Generic;

namespace GestureToolkit
{
    /// <summary>
    /// Implements input / output operations for gesture clusters
    /// </summary>
    public class ClusterIO
    {
        /// <summary>
        /// Writes clusters to file
        /// </summary>
        /// <param name="fileName"></param>
        public static void WriteClustersToFile(string fileName, List<Gesture> gestures)
        {
            using (StreamWriter writer = new StreamWriter(fileName))
            {
                // write header line with column names
                writer.WriteLine("participant,gesture,numberOfStrokes,clusterName,gestureFileName");

                // write a single line for each gesture
                for (int i = 0; i < gestures.Count; i++)
                {
                    Gesture gesture = gestures[i];
                    writer.WriteLine(
                        "{0},{1},{2},{3},{4}",
                        gesture.ParticipantName,
                        gesture.Name,
                        gesture.GetNumberOfStrokes(),
                        gesture.ClusterName,
                        gesture.FileName
                    );
                }
            }
        }

        /// <summary>
        /// Computes agreement rates from a cluster file 
        /// </summary>
        /// <param name="clusterFileName"></param>
        /// <param name="ouputFileName"></param>
        public static void ComputeAgreementRates(string clusterFileName, string outputFileName, string participant = "")
        {
            // associates gesture names (key) with a list of clusters (value)
            // the list of clusters is also a dictionary that associates cluster names (key) with cluster size (value)
            Dictionary<string, Dictionary<string, int>> dictionary = new Dictionary<string, Dictionary<string, int>>();
            using (StreamReader reader = new StreamReader(clusterFileName))
            {
                string line = reader.ReadLine(); // first line contains column names and therefore ignore
                while ((line = reader.ReadLine()) != null)
                {
                    string[] values = line.Split(new char[] { ',' });
                    string gestureName = values[1].Split(new char[] { '~' })[0];
                    string clusterName = values[3];

                    if (!dictionary.ContainsKey(gestureName))
                        dictionary.Add(gestureName, new Dictionary<string, int>());
                    if (!dictionary[gestureName].ContainsKey(clusterName))
                        dictionary[gestureName].Add(clusterName, 0);
                    dictionary[gestureName][clusterName]++;
                }
            }

            // compute agreement rates
            using (StreamWriter writer = new StreamWriter(outputFileName))
            {
                writer.WriteLine("participant,gesture,agreementRate(2005_Formula),agreementRate(2015_Formula)");
                foreach (KeyValuePair<string, Dictionary<string, int>> gestureClusters in dictionary)
                {
                    double numSamples = 0;
                    double agreementRate_2005 = 0;
                    foreach (KeyValuePair<string, int> cluster in gestureClusters.Value)
                    {
                        numSamples += cluster.Value;
                        agreementRate_2005 += cluster.Value * cluster.Value;
                    }
                    agreementRate_2005 /= (numSamples * numSamples);

                    double agreementRate_2015 = (numSamples * agreementRate_2005 - 1) / (numSamples - 1);
                    writer.WriteLine("{0},{1},{2:.0000},{3:.0000}", participant, gestureClusters.Key, agreementRate_2005, agreementRate_2015);
                }

                writer.WriteLine();
                writer.WriteLine("\"agreementRate(2005_Formula) reports agreement rates using the formula from:\"");
                writer.WriteLine("\"Jacob O. Wobbrock, Htet Htet Aung, Brandon Rothrock, and Brad A. Myers. 2005. Maximizing the guessability of symbolic input. In CHI '05 Extended Abstracts on Human Factors in Computing Systems (CHI EA '05). ACM, New York, NY, USA, 1869-1872.\"");
                writer.WriteLine();
                writer.WriteLine("\"agreementRate(2015_Formula) reports agreement rates using the updated formula from:\"");
                writer.WriteLine("\"Radu-Daniel Vatavu and Jacob O. Wobbrock. 2015. Formalizing Agreement Analysis for Elicitation Studies: New Measures, Significance Test, and Toolkit. In Proceedings of the 33rd Annual ACM Conference on Human Factors in Computing Systems (CHI '15). ACM, New York, NY, USA, 1325-1334.\"");
            }
        }
    }
}
