﻿/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.IO;

namespace GestureToolkit
{
    /// <summary>
    /// Implements hierarchical clustering of a gesture set
    /// </summary>
    public class HierarchicalClustering
    {
        /// <summary>
        /// Performs hierarchical clustering on the set of gestures provided as the argument.
        /// Returns a list of clusters.
        /// </summary>
        /// <param name="gestures"></param>
        public List<Cluster> Cluster(Gesture[] gestures, float minSimilarityThreshold)
        {
            // compute similarity matrix
            float[][] similarity = ComputeSimilarityMatrix(gestures);
            
            // initialize clusters: each cluster contains exactly one gesture sample
            List<Cluster> clusters = new List<Cluster>();
            for (int i = 0; i < gestures.Length; i++)
            {
                Cluster cluster = new Cluster();
                cluster.Add(i);
                clusters.Add(cluster);
            }
            
            // perform hierarchical clustering:
            // clusters are merged while the percentage of similarity remains above threshold
            while (clusters.Count > 1)
            {
                // find two clusters to merge
                float maxSimilarity = float.MinValue;
                int index1 = -1, index2 = -1;
                for (int i = 0; i < clusters.Count; i++)
                    for (int j = i + 1; j < clusters.Count; j++)
                    {
                        float sim = ComputeSimilarityBetweenClusters(gestures, clusters[i], clusters[j], similarity);
                        if (sim > maxSimilarity)
                        {
                            maxSimilarity = sim;
                            index1 = i;
                            index2 = j;
                        }
                    }
                
                // decide whether to merge clusters or stop the clustering process
                if (maxSimilarity >= minSimilarityThreshold)
                {
                    clusters[index1].AddRange(clusters[index2]);
                    clusters[index1].Sort(); 
                    clusters.RemoveAt(index2);
                }
                else break; // can't merge clusters any longer as they are too *far* apart given the minSimilarityThreshold
            }

            // report clusters
            return clusters;
        }

        /// <summary>
        /// Computes the similarity matrix for all pairs of gestures (to speed up computations)
        /// </summary>
        /// <param name="gestures"></param>
        /// <returns></returns>
        private float[][] ComputeSimilarityMatrix(Gesture[] gestures)
        {
            // shortcut notation for the number of gestures
            int n = gestures.Length;

            // get number of strokes for each gesture (needed for similarity computation)
            int[] numStrokesPerGesture = new int[n];
            for (int i = 0; i < n; i++)
                numStrokesPerGesture[i] = gestures[i].GetNumberOfStrokes();

            // compute similarity matrix
            float[][] similarity = new float[n][];
            for (int i = 0; i < n; i++)
                similarity[i] = new float[n];
            for (int i = 0; i < n; i++)
            {
                similarity[i][i] = 1.0f;
                for (int j = i + 1; j < n; j++)
                {
                    similarity[i][j] = ComputeSimilarityBetweenGestures(gestures[i], gestures[j], numStrokesPerGesture[i], numStrokesPerGesture[j]);
                    similarity[j][i] = similarity[i][j];
                }
            }

            return similarity;
        }

        /// <summary>
        /// Computes the similarity between two gestures with value in [0..1],
        /// with 1 denoting greatest similarity (perfect match).
        /// Employs the Euclidean distance.
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        private float ComputeSimilarityBetweenGestures(Gesture a, Gesture b, int numStrokesA, int numStrokesB)
        {
            // same type?
            if (a.Name.CompareTo(b.Name) != 0)
                return 0.0f;

            // same number of strokes?
            if (numStrokesA != numStrokesB)
                return 0.0f;

            return 1.0f - GestureMetrics.Euclidean(a, b) / a.SampledPoints.Length;
        }

        /// <summary>
        /// Computes the similarity between two clusters as the minimum similarity of their members
        /// </summary>
        /// <param name="gestures"></param>
        /// <param name="cluster1"></param>
        /// <param name="cluster2"></param>
        /// <param name="similarity"></param>
        /// <returns></returns>
        private float ComputeSimilarityBetweenClusters(
            Gesture[] gestures, 
            Cluster cluster1, Cluster cluster2, 
            float[][] similarity)
        {
            float sim = float.MaxValue;

            foreach (int index1 in cluster1)
                foreach (int index2 in cluster2)
                    sim = Math.Min(sim, similarity[index1][index2]);

            return sim;
        }
    }
}