/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System.Collections.Generic;

namespace GestureToolkit
{
    public class Gesture
    {
        private const int SAMPLING_RESOLUTION = 64;

        public Point[] m_normalizedPoints = null;   // raw normalized motion trajectory
        public Point[] SampledPoints = null;        // sampled points 
        public int[] m_strokeIndex = null;          // m_strokeIndex[i] represents the stroke index for point i

        public string Name = "";                    // gesture name, e.g. "arrowright"
        public string ParticipantName = "";         // participant from which this gesture sample came from (usually the name of the folder containing gesture samples, e.g. "11")
        public string FileName = "";                // file name from which this gesture was loaded, e.g. "arrowright~05.xml". Acts as a unique identifier for this gesture.
        public string ClusterName = "";             // cluster name resulted from the clustering process, e.g. "5" (cluster number 5)

        /// <summary>
        /// Constructs a gesture from a collection of points
        /// </summary>
        public Gesture(Point[] points, int[] strokeIndex)
        {
            // store stroke information for each point
            this.m_strokeIndex = strokeIndex;
            
            // normalize points without distortion in the [-1,1]^2 interval 
            // and resample path
            if (points != null)
            {
                m_normalizedPoints = PointsPreprocessing.Normalize(points);
                SampleUniformly(SAMPLING_RESOLUTION);
            }
        }

        /// <summary>
        /// Samples the motion trajectory with the specified resolution
        /// </summary>
        public void SampleUniformly(int numPoints)
        {
            SampledPoints = PointsPreprocessing.SampleEqualLength(m_normalizedPoints, m_strokeIndex, numPoints - 1);
        }

        /// <summary>
        /// Returns the number of individual strokes
        /// </summary>
        /// <returns></returns>
        public int GetNumberOfStrokes()
        {
            int numStrokes = 1;
            for (int i = 1; i < m_strokeIndex.Length; i++)
                if (m_strokeIndex[i] != m_strokeIndex[i - 1])
                    numStrokes++;
            return numStrokes;
        }
    }
}