/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System.IO;
using System.Collections.Generic;
using System.Xml;

namespace GestureToolkit
{
    public class GestureIO
    {
        /// <summary>
        /// Reads a gesture from a file (one gesture per file)
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public static Gesture ReadGesture(string fileName)
        {
            List<Point> points = new List<Point>();
            List<int> strokeIndex = new List<int>();

            // read strokes and points
            XmlTextReader xmlReader = null;
            float startTime = 0;
            int currentStrokeIndex = -1;
            string gestureName = "";
            string participantName = "";
            try
            {
                xmlReader = new XmlTextReader(File.OpenText(fileName));
                while (xmlReader.Read())
                {
                    if (xmlReader.NodeType != XmlNodeType.Element) continue;
                    switch (xmlReader.Name)
                    {
                        case "Gesture":
                            gestureName = xmlReader["Name"];
                            participantName = xmlReader["Subject"];
                            break;
                        case "Stroke":
                            currentStrokeIndex++;
                            break;
                        case "Point":
                            if (points.Count == 0)
                                startTime = float.Parse(xmlReader["T"]);
                            float time = 0;
                            try
                            {
                                time = float.Parse(xmlReader["T"]) - startTime; // some sets (such as NicIcon) don't have valid time values
                            }
                            catch { }
                            points.Add(new Point(
                                float.Parse(xmlReader["X"]),
                                float.Parse(xmlReader["Y"]),
                                0,
                                time
                            ));
                            strokeIndex.Add(currentStrokeIndex);
                            break;
                    }
                }
            }
            finally
            {
                if (xmlReader != null)
                    xmlReader.Close();
            }

            // create gesture
            Gesture gesture = new Gesture(points.ToArray(), strokeIndex.ToArray());
            
            // store gesture name
            if (gestureName.Length == 0)
            {
                // extract gesture name from fileName, e.g. "arrowright" from "arrowright~05.xml"
                gestureName = fileName.Substring(fileName.LastIndexOf("\\") + 1);
                gestureName = gestureName.Substring(0, gestureName.Length - 4);
            }
            if (gestureName.IndexOf('~') >= 0)
                gestureName = gestureName.Substring(0, gestureName.IndexOf('~'));  // remove the repetition number (e.g., arrow~09 indicates the 9th repetition of the arrow gesture in the MMG set)
            gesture.Name = gestureName;

            // store participant name
            if (participantName.Length == 0)
            {
                // extract participant name from folder name, e.g. "11" from "11\arrowright~05.xml"
                participantName = fileName.Substring(0, fileName.LastIndexOf("\\"));
                participantName = participantName.Substring(participantName.LastIndexOf("\\") + 1);
            }
            gesture.ParticipantName = participantName;

            // store gesture (short) file name, e.g. "arrowright~05.xml"
            gesture.FileName = (new FileInfo(fileName)).Name;
            
            return gesture;
        }
    }
}
