/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;

namespace GestureToolkit
{
	public class Point
	{
		public float X, Y, Z;      // coordinates on 3 axis 
        public float T;            // time [in milliseconds]

		#region constructors, clone

		/// <summary>
		/// Creates a point with (0, 0, 0, 0) coordinates
		/// </summary>
		public Point()
		{	
			this.LoadMemberData(0, 0, 0, 0);
		}

		/// <summary>
		/// Creates a 2D point with (x, y, 0, 0) coordinates
		/// </summary>
        public Point(float x, float y)
		{
			this.LoadMemberData(x, y, 0, 0);
		}

		/// <summary>
		/// Creates a 3D point with (x, y, z, 0) coordinates
		/// </summary>
        public Point(float x, float y, float z)
		{
			this.LoadMemberData(x, y, z, 0);
		}

        /// <summary>
        /// Creates a 4D point with (x, y, z, t) coordinates
        /// </summary>
        public Point(float x, float y, float z, float t)
        {
            this.LoadMemberData(x, y, z, t);
        }

		/// <summary>
		/// Creates a point identical to point p
		/// </summary>
		public Point(Point p)
		{
			this.LoadMemberData(p.X, p.Y, p.Z, p.T);
		}
		
		/// <summary>
		/// Creates a point identical with this point
		/// </summary>
		public Point Clone()
		{
            return new Point(this);
		}

        private void LoadMemberData(float x, float y, float z, float t)
		{
			X = x;
			Y = y;
			Z = z;
            T = t;
		}

		#endregion

		#region special points: zero

		/// <summary>
		/// Returns the point with coordinates (0, 0, 0, 0)
		/// </summary>
		public static Point Zero
		{
			get
			{
				return new Point(0, 0, 0, 0);
			}
		}

		#endregion

        #region properties: max coordinate

        /// <summary>
		/// Returns the max value of the coordinates = max(x, y, z)
		/// </summary>
        public float MaxCoordinate
		{
			get
			{
				return Math.Max(X, Math.Max(Y, Z));
			}
		}

		#endregion

		#region operations: +, -, *, /

		/// <summary>
		/// Returns p1 + p2 on each coordinate
		/// </summary>
		public static Point operator +(Point p1, Point p2)
		{
			return new Point(p1.X + p2.X, p1.Y + p2.Y, p1.Z + p2.Z);
		}

		/// <summary>
		/// Returns p1 - p2 on each coordinate
		/// </summary>
		public static Point operator -(Point p1, Point p2)
		{
			return new Point(p1.X - p2.X, p1.Y - p2.Y, p1.Z - p2.Z);
		}

		/// <summary>
		/// Returns p + scalar on each coordinate
		/// </summary>
        public static Point operator +(Point p, float scalar)
		{
			return new Point(p.X + scalar, p.Y + scalar, p.Z + scalar, p.T);
		}
	
		/// <summary>
		/// Returns p / scalar on each coordinate
        /// Throws DivideByZeroException if abs(scalar) is less than 10e-6
		/// </summary>
        public static Point operator /(Point p, float scalar)
		{
            if (Math.Abs(scalar) < 10e-6)
                throw new DivideByZeroException(String.Format("scalar is less than 10e-6 [{0}]", scalar));
			return new Point(p.X / scalar, p.Y / scalar, p.Z / scalar, p.T);
		}

		/// <summary>
		/// Returns scalar * p on each coordinate
		/// </summary>
        public static Point operator *(float scalar, Point p)
		{
			return new Point(scalar * p.X, scalar * p.Y, scalar * p.Z, p.T);
		}

       	#endregion

		#region ToString

		/// <summary>
		/// Returns a string representation of this point with 3 decimals per coordinate:
		/// ex: 2.314 7.000 10.736
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			return String.Format("{0:.000} {1:.000} {2:.000}", X, Y, Z);
		}

		#endregion
	}
}
