/**
 * The GEsture Clustering ToolKit (GECKo) (.NET Framework 4.0 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 *	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for GECKo, and what should be 
 * used to cite it, is:
 *
 *	Anthony, L., Vatavu, R.-D., and Wobbrock, J.O. (2013).  
 *	  Understanding the Consistency of Users' Pen and Finger Stroke Gesture Articulation. 
 *	  Proceedings of the 39th Graphics Interface Conference - GI'13 
 *	  (Regina, Saskatchewan, Canada, May 2013). Toronto, Ontario: Canadian Information Processing Society.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
namespace GestureToolkit
{
    public class PointsPreprocessing
    {
        /// <summary>
        /// Normalizes an array of points with shape preservation so that they fall in [-1,1]^2
        /// New center of gravity becomes (0, 0)
        /// Complexity: O(n)
        /// </summary>
        public static Point[] Normalize(Point[] points)
        {
            if (points == null || points.Length == 0)
                return new Point[0];

            // new points will be computed during normalization
            Point[] newPoints = new Point[points.Length];

            // compute min and max
            Point min = new Point(points[0]);
            Point max = new Point(points[0]);
            for (int i = 1; i < points.Length; i++)
            {
                Point current = points[i];
                if (min.X > current.X) min.X = current.X;
                if (min.Y > current.Y) min.Y = current.Y;
                if (max.X < current.X) max.X = current.X;
                if (max.Y < current.Y) max.Y = current.Y;
            }

            // scale points to [0,1]^2
            float scale = 1.0f / (max - min).MaxCoordinate;
            for (int i = 0; i < points.Length; i++)
                newPoints[i] = scale * (points[i] - min);

            // set new center of gravity to (0, 0) for newPoints
            // which means that the new (normalized) points will fall into [-1,1]^2
            Point cg = new Point(0, 0);
            for (int i = 0; i < newPoints.Length; i++)
                cg += newPoints[i];
            cg /= newPoints.Length;
            for (int i = 0; i < newPoints.Length; i++)
                newPoints[i] -= cg;

            return newPoints;
        }

        /// <summary>
        /// Samples an array of points into a specified number of segments of equal length (uniform sampling)
        /// Complexity: O(n)
        /// </summary>
        public static Point[] SampleEqualLength(Point[] points, int[] strokeIndex, int numSegments)
        {
            // compute total length and segment length
            float totalLength = GetPolyLineLength(points, strokeIndex);
            float I = totalLength / numSegments;

            // allocate memory for the array of simplified points 
            // and add the first point
            Point[] newPoints = new Point[numSegments + 1];
            int cntNewPoints = 0;
            newPoints[cntNewPoints++] = points[0].Clone();
            
            // add rest of the points
            float D = 0;
            for (int i = 1; i < points.Length; i++)
            {
                //if (strokeIndex[i] != strokeIndex[i - 1]) continue;
                
                float d = Measurements.EuclideanDistance(points[i - 1], points[i]);
                if (D + d < I)
                    D += d;
                else
                {
                    Point firstPoint = points[i - 1];
                    while (D + d >= I)
                    {
                        // add interpolated point
                        float t = (I - D) / d;
                        if (t < 0.0) t = 0.0f;
                        if (t > 1.0) t = 1.0f;
                        newPoints[cntNewPoints++] = (1.0f - t) * firstPoint + t * points[i];
                        
                        // update partial length
                        d = D + d - I;
                        D = 0; 
                        firstPoint = newPoints[cntNewPoints - 1];
                    }
                    D = d;
                }
            }

            if (cntNewPoints == numSegments + 1)
                cntNewPoints--;
            newPoints[cntNewPoints++] = points[points.Length - 1].Clone();
            
            return newPoints;
        }

        /// <summary>
        /// Computes the length of the polyline given by the array of points
        /// Complexity: O(n)
        /// </summary>
        public static float GetPolyLineLength(Point[] points, int[] strokeIndex)
        {
            float length = 0;
            for (int i = 1; i < points.Length; i++)
                //if (strokeIndex[i] == strokeIndex[i - 1])
                    length += Measurements.EuclideanDistance(points[i - 1], points[i]);
            return length;
        }
    }
}
